# Atomic Force Drops and Snaps (AFDS)
# 9/2006 based on earlier CSDrops and FindJumps
# * lists the "Drops" characterized by piezo moving down
# * lists the "Snaps" -- sudden changes in force when the piezo is rising
#   - Snap detection is rudimentary -- perhaps Markov/Kalman methods would be more robust
# * makes a Result of the Snaps, with stats about each snap and its trace
#   - "Trace Idx" -- which sel it came from
#   - "Trace Min" -- min force whole trace
#   - "Trace Max" -- max force whole trace
#   - "Trace Std" -- rms noise of first k points
#   - "Rel t"     -- delta time since piezo change direction
#   - "Rel F"     -- delta force " " "
#   - "Delta t extrema" -- delta time from top to bottom of snap
#   - "Delta F extrema" -- delta force  " " "
#   - "Delta F context" -- delta force between pre-context mean and post-context mean
#   - "Delta F detect"  -- detection delta force: the jump is found at sample i which maximizes: y[i+deltaP] - y[i]
#     (this is a cruder measurement than any of the above)
#   - "PreCtx Mean"     -- pre-context mean
#   - "PostCtx Mean"    -- post-context mean
# * included WordReport plugin outputs Snaps as an Excel sheet

import QUB
from Buttons import *
from Dialogs import *
from qubtree import Node, NullNode
from qubtree import QTR_TYPE_FLOAT
from math import *
from itertools import *
import Reports
import WordReport


# utility

def PutAll(putter):
    return putter
def PutAlt(putter, initOn):
    on = initOn
    def Put(*a, **kw):
        if on:
            putter(*a, **kw)
        on = (on+1) % 2
    return Put
def PutOdds(putter):
    return PutAlt(putter, 1)
def PutEvens(putter):
    return PutAlt(putter, 0)

def children(tree, name=None):
    if name == None:
        child = tree.child
    else:
        child = tree.find(name)
    while not child.isNull:
        yield child
        if name == None:
            child = child.sibling
        else:
            child = child.nextSameName()

def GenDataSegments(dataset, channel=-1):
    ds = dataset.clone()
    if channel >= 0:
        ds['ActiveChannel'].data = channel
    for seg in children(ds, 'Segment'):
        QUB.fillData(seg)
        yield seg
        seg.remove(seg['Channel'])

# find drops by listing where the piezo channel is {'increasing':1, 'decreasing':-1}

def GenSelsBySlope(data, first, direction):
    if direction < 0:
        isDir = lambda x, y: x > y
    else:
        isDir = lambda x, y: x < y
    last = None
    for i, x in enumerate(data):
        if last == None:
            spot = 0
        elif not isDir(last, x):
            if spot != (i-1):
                yield (first+spot, first+i-1)
            spot = i
        last = x
    if spot != (len(data) - 1):
        yield (first+spot, first + len(data) - 1)

                
# find snaps by taking a lame derivative: data[i+deltaP] - data[i]
# then finding peaks in the derivative

def Derivate(lst, deltaP, pcter=lambda x:x):
    '''returns [ lst[i+deltaP] - lst[i] ]'''
    if len(lst) <= deltaP:
        return []
    der = []
    nn = 1.0 * len(lst)
    for i in range(len(lst) - deltaP):
        der.append( lst[i+deltaP] - lst[i] )
        if i%50 == 0:
            pcter(i/nn)
    return der

def FindPeaks(lst, minPeak, skip, pcter=lambda x:x):
    peaks = []
    if len(lst) <= 1:
        return peaks

    nn = len(lst)    
    if lst[0] > lst[1] and lst[0] >= minPeak:
        peaks.append( (0, lst[0]) )
    for i in range(1, len(lst)-1):
        if (not len(peaks)) or ((i - peaks[len(peaks)-1][0]) > skip):
            v = lst[i]
            if v > lst[i-1] and v >= lst[i+1] and v >= minPeak:
                peaks.append( (i, v) )
        if i%50 == 0:
            pcter(i/nn)
    i = len(lst) - 1
    if lst[i] > lst[i-1] and lst[i] >= minPeak:
        peaks.append( (i, lst[i]) )
    return peaks            
    
def FindNegPeaks(lst, minPeak, skip, pcter=lambda x:x):
    peaks = []
    if len(lst) <= 1:
        return peaks

    nn = len(lst)    
    if lst[0] < lst[1] and lst[0] <= minPeak:
        peaks.append( (0, lst[0]) )
    for i in range(1, len(lst)-1):
        if (not len(peaks)) or ((i - peaks[len(peaks)-1][0]) > skip):
            v = lst[i]
            if v < lst[i-1] and v <= lst[i+1] and v <= minPeak:
                peaks.append( (i, v) )
        if i%50 == 0:
            pcter(i/nn)
    i = len(lst) - 1
    if lst[i] < lst[i-1] and lst[i] <= minPeak:
        peaks.append( (i, lst[i]) )

    return peaks            
    
def AvgContext(seg, p0, deltaP, contextPreP, contextPostP):
    def Avg(ixs):
        sum, count = 0.0, 0
        for ix in ixs:
            sum += data[ix]
            count += 1
        if count > 0:
            sum /= count
        return sum
            
    data = seg['Channel'].data
    N = len(data)

    return (Avg(xrange(max(0, p0-contextPreP), max(0, p0))),
            Avg(xrange(min(N, p0+deltaP+1), min(N, p0+deltaP+contextPostP+1))))

def SnapStat(fdata, i, deltaP):
    minf, mini = fdata[i], i
    maxf, maxi = fdata[i], i
    for ii in xrange(i+1, i+deltaP+1):
        if fdata[ii] < minf:
            minf, mini = fdata[ii], ii
        if fdata[ii] > maxf:
            maxf, maxi = fdata[ii], ii
    return minf, maxf, abs(maxi - mini)

def TraceStat(fdata, rmsSampleP):
    i0 = min(len(fdata)-rmsSampleP, rmsSampleP)
    n = min(len(fdata), rmsSampleP)
    fmin, fmax = fdata[i0], fdata[i0]
    fsum = 0.0
    for i, x in enumerate(fdata):
        if x < fmin:
            fmin = x
        if x > fmax:
            fmax = x
        if i0 <= i < (i0+n):
            fsum += x
    fmean = fsum / n
    fsum = 0.0
    for i in xrange(i0, i0+n):
        fsum += (fmean - fdata[i]) * (fmean - fdata[i])
    return fmin, fmax, sqrt(fsum / n)


def AFDS(props):
    # data
    dataFile = QUB.DataView.currData
    data = QUB.getChosenData()
    dataset = data.find('DataSet')
    sampling = dataset['sampling'].data[0]
    TtoP = lambda t: int(round(t / sampling))

    # drop setup and storage
    droplist = []
    ptsThresh = props['ptsThresh']
    def PDrop(drop):
        f, l = drop[0], drop[1]
        if (l - f + 1) >= ptsThresh:
            droplist.append(drop)
    PutDrop = {'all'   : PutAll(PDrop),
               'odds'  : PutOdds(PDrop),
               'evens' : PutEvens(PDrop) }[ props['keep'] ]
    direction = {'increasing':1, 'decreasing':-1}[ props["PiezoDir"] ]

    # snap setup
    contextPreP = TtoP(props['contextPre'])
    contextPostP = TtoP(props['contextPost'])
    rmsSampleP = TtoP(props['rmsSample'])
    deltaY = props['deltaY']
    deltaT = props['deltaT']
    deltaP = int(round(deltaT / sampling))
    #----- Point XFindPeaks to positive or negative peaks function 
    XFindPeaks = FindPeaks
    if deltaY < 0:
        XFindPeaks = FindNegPeaks
    optionalContextP = 0
    if props['skipContext'] == 'yes':
        optionalContextP = contextPostP

    # snap storage and results
    result = Node('Results')
    result['ResultType'].data = 'AFDS'
    result['DataSet']['FileName'].data = str(dataset['FileName'].data)
    rsegs = result['DataSet']
    rsegb = [NullNode()] #boxed so we can assign it from PutSnap
    snaplist = []
    def PutSnap(snap_i, fseg, top_i, trace_i, snap_idx, trace_min, trace_max, trace_rms):
        fdata = fseg['Channel'].data
        trace_f = fseg.data[0]
        trace_l = fseg.data[1]
        first = max(trace_f, snap_i - contextPreP)
        last = min(trace_l, snap_i + deltaP + contextPostP - 1)
        snaplist.append( (first, last) )
        rsegb[0] = rsegs.insert('Segment', rsegb[0])
        rseg = rsegb[0]
        rseg.data = (first, last)
        rseg['start'].data = fseg['start'].data[0] + sampling * (first - trace_f)
        rseg['length'].data = (last - first + 1) * sampling
        rseg['Trace Idx'].data = trace_i
        rseg['Snap Idx'].data = snap_idx
        rseg['Trace Min'].data = trace_min
        rseg['Trace Max'].data = trace_max
        rseg['Trace Std'].data = trace_rms
        rseg['Rel t'].data = sampling * (snap_i - top_i)
        rseg['Rel F'].data = fdata[snap_i-trace_f] - fdata[top_i-trace_f]
        minf, maxf, sam = SnapStat(fdata, snap_i-trace_f, deltaP)
        rseg['Delta t extrema'].data = sampling * sam
        rseg['Delta F extrema'].data = maxf - minf
        rseg['PreCtx Mean'].data, rseg['PostCtx Mean'].data = AvgContext(fseg, snap_i-trace_f, deltaP, contextPreP, contextPostP)
        rseg['Delta F context'].data = rseg['PostCtx Mean'].data[0] - rseg['PreCtx Mean'].data[0]
        rseg['Delta F detect'].data = fdata[snap_i-trace_f + deltaP] - fdata[snap_i-trace_f]
        
        
    # iterate over segments,
    piezosegs = GenDataSegments(dataset, props['ChnlPiezo'])
    forcesegs = GenDataSegments(dataset, props['ChnlForce'])
    trace_i = 0
    for pseg, fseg in izip(piezosegs, forcesegs):
        # putting all drops
        f, l = pseg.data[0], pseg.data[1]
        drops = [drop for drop in GenSelsBySlope(pseg['Channel'].data, f, direction)]
        if len(drops) == 0:
            continue
        for i in xrange(len(drops)):
            PutDrop(drops[i])
        drop = drops[0]
        pmaxi = drop[1]
        pmaxy = pseg['Channel'].data[pmaxi-f]

        # and snaps
        trace_min, trace_max, trace_rms = TraceStat(fseg['Channel'].data, rmsSampleP)
        deriv = Derivate(fseg['Channel'].data[pmaxi-f+1:], deltaP)
        peaks = XFindPeaks(deriv, deltaY, deltaP + optionalContextP)
        snap_idx = 0
        for p_i, p_y in peaks:
            PutSnap(p_i+pmaxi+1, fseg, pmaxi, trace_i, snap_idx, trace_min, trace_max, trace_rms)
            snap_idx += 1
        trace_i += 1
                

    # shorten the drops to make them the same length
    if len(droplist) and props['Trim'] != 'neither':
        N = min([l-f+1 for f,l in droplist])
        if props['Trim'] == 'left':
            Trim = lambda f, l: (f + ((l-f+1)-N), l)
        else:
            Trim = lambda f, l: (f, l - ((l-f+1)-N))
        droplist = [Trim(f, l) for f,l in droplist]
        
    dataFile.setList(props['DropListName'], droplist)
    dataFile.setList(props['SnapListName'], snaplist)
    QUB.ResultsView.addResults( result )


AddTaskButton( 'AFDS', # button text
               AFDS,   # function
               'AFDS',       # short text for small buttons
               0,            # show settings dialog on left-click?
               1,            # show settings dialog on right-click?
               "Atomic Force Drops and Snaps",   # tool-tip
               "Atomic Force Drops and Snaps",   # description in "Edit Buttons"
               "AFDS settings:", # caption for the settings dialog
               [DialogText("ChnlPiezo", "Piezo Channel:", 1, acceptIntGreaterThan(-1)),
                DialogText("DropListName", "Drop List name:", "Drops", acceptString),
                DialogRadioGroup("PiezoDir", "List where piezo is", ['increasing', 'decreasing'], 'increasing'),
                DialogRadioGroup("keep", 'Keep drops:', ['all', 'odds', 'evens'], 'all'),
                DialogText("ptsThresh", 'Ignore sels shorter than (points):',400, acceptIntGreaterThan(0)),
                DialogRadioGroup("Trim", 'For uniform width, trim', ['left', 'right', 'neither'], 'left'),
                
                DialogText("ChnlForce", "Force Channel:", 0, acceptIntGreaterThan(-1)),
                DialogText("SnapListName", "Snap List name:", "Snaps", acceptString),
                DialogText("deltaT", "Delta t (ms):", 0.5, acceptFloatGreaterThan(0.0)),
                DialogText("deltaY", "Min delta y:", 1.0, acceptFloat),
                DialogText("contextPre", "Context pre (ms):", 0.5, acceptFloatGreaterThan(-0.00001)),
                DialogText("contextPost", "Context post (ms):", 0.5, acceptFloatGreaterThan(-0.00001)),
                DialogRadioGroup("skipContext", "Skip Context:", ['yes', 'no'], 'yes'),
                DialogText("msecSkip", "Skip initial msec's:", 0.0, acceptFloatGreaterThan(-0.00001)),
                DialogText("rmsSample", "RMS Sample (ms):", 5.0, acceptFloatGreaterThan(0.0))],
               'Preprocessing' )


def ReportSnaps(results, context):
    rep = WordReport.EnsureWordReport(context)
    node = results.node

    hdr = ['start', 'length', 'Trace Idx', 'Snap Idx', 'Trace Min', 'Trace Max', 'Trace Std',
           'Rel t', 'Rel F', 'Delta t extrema', 'Delta F extrema', 'Delta F context', 'Delta F detect',
           'PreCtx Mean', 'PostCtx Mean']
    tbl = [ hdr ]

    for seg in children(node['DataSet'], 'Segment'):
        tbl.append( map( lambda field: seg[field].data[0], hdr ) )
        
    rep.writeTable(tbl)    

Reports.Add('Snaps', ReportSnaps, ['AFDS'])
